/*
 * Copyright (c) 2007 Henri Sivonen
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

package fi.iki.hsivonen.bib4ht;

import java.util.Map;

import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLFilterImpl;

import fi.iki.hsivonen.xml.AttributesImpl;

public class AbbrFilter extends XMLFilterImpl {
    
    private Map<String, String> definitions;

    private StringBuilder buffer = new StringBuilder();

    private boolean inAbbr;

    private Locator locator;

    private boolean prevWasWhitespace;
    
    private AttributesImpl attrs;
    
    public AbbrFilter(XMLReader parent, Map<String, String> definitions) {
        super(parent);
        this.definitions = definitions;
    }

    /**
     * @see org.xml.sax.helpers.XMLFilterImpl#startDocument()
     */
    @Override
    public void startDocument() throws SAXException {
        inAbbr = false;
        super.startDocument();
    }

    /**
     * @see org.xml.sax.helpers.XMLFilterImpl#characters(char[], int, int)
     */
    @Override
    public void characters(char[] ch, int start, int length) throws SAXException {
        if (inAbbr) {
            for (int i = start; i < start + length; i++) {
                char c = ch[i];
                if (c == ' ' || c == '\t' || c == '\n' || c == '\r') {
                    if (!prevWasWhitespace) {
                        buffer.append(' ');
                        prevWasWhitespace = true;
                    }
                } else {
                    buffer.append(c);
                    prevWasWhitespace = false;
                }
            }
        } else {
            super.characters(ch, start, length);
        }
    }


    /**
     * @see org.xml.sax.helpers.XMLFilterImpl#endElement(java.lang.String, java.lang.String, java.lang.String)
     */
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        if ("http://www.w3.org/1999/xhtml".equals(uri) && "abbr".equals(localName)) {
            int len = buffer.length() - 1;
            if (len >= 0 && buffer.charAt(len) == ' ') {
                buffer.setLength(len);
            }
            String abbr = buffer.toString();
            if (attrs.getIndex("", "title") < 0) {
                String definition = definitions.get(abbr);
                if (definition == null) {
                    error(new SAXParseException("No definition for " + abbr, locator));
                } else {
                    attrs.addAttribute("title", definition);
                }
            }
            super.startElement(uri, localName, qName, attrs);
            super.characters(abbr.toCharArray(), 0, abbr.length());
            super.endElement(uri, localName, qName);
            inAbbr = false;
        } else {
            super.endElement(uri, localName, qName);
        }
    }


    /**
     * @see org.xml.sax.helpers.XMLFilterImpl#setDocumentLocator(org.xml.sax.Locator)
     */
    @Override
    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
        super.setDocumentLocator(locator);
    }


    /**
     * @see org.xml.sax.helpers.XMLFilterImpl#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
     */
    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        if (inAbbr) {
            SAXParseException ex = new SAXParseException("Child element in <abbr>!", locator);
            fatalError(ex);
            throw ex;
        }
        if ("http://www.w3.org/1999/xhtml".equals(uri) && "abbr".equals(localName)) {
            attrs = new AttributesImpl(atts);
            inAbbr = true;
            prevWasWhitespace = true;
            buffer.setLength(0);
        } else {
            super.startElement(uri, localName, qName, atts);
        }
    }
    
    
}
