/*
 * Copyright (c) 2007 Henri Sivonen
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

package fi.iki.hsivonen.bib4ht;

import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

public class DefinitionGatherer extends DefaultHandler {

    private static final Pattern GLOSSARY = Pattern.compile(".*\\bglossary\\b.*");

    private enum ParseState {
        OUTSIDE_GLOSSARY, IN_GLOSSARY, IN_GLOSSARY_TERM, IN_GLOSSARY_DEFINITION, IN_GLOSSARY_DEFINITION_SPAN
    }

    private StringBuilder buffer = new StringBuilder();

    private Map<String, String> definitions;

    private ParseState state;

    private int spanEndsToIgnore;

    private String term;

    private Locator locator;

    private boolean prevWasWhitespace;

    /**
     * @see org.xml.sax.helpers.DefaultHandler#startDocument()
     */
    @Override
    public void startDocument() throws SAXException {
        state = ParseState.OUTSIDE_GLOSSARY;
        definitions = new HashMap<String, String>();
        spanEndsToIgnore = 0;
        term = null;
    }

    /**
     * @see org.xml.sax.helpers.DefaultHandler#startElement(java.lang.String,
     *      java.lang.String, java.lang.String, org.xml.sax.Attributes)
     */
    @Override
    public void startElement(String uri, String localName, String qName,
            Attributes atts) throws SAXException {
        if ("http://www.w3.org/1999/xhtml".equals(uri)) {
            switch (state) {
                case OUTSIDE_GLOSSARY:
                    if ("dl".equals(localName)) {
                        String klazz = atts.getValue("", "class");
                        if (klazz != null) {
                            Matcher m = GLOSSARY.matcher(klazz);
                            if (m.matches()) {
                                state = ParseState.IN_GLOSSARY;
                            }
                        }
                    }
                    break;
                case IN_GLOSSARY:
                    if ("dl".equals(localName)) {
                        throw new SAXParseException("<dl> inside glossary.",
                                locator);
                    } else if ("span".equals(localName)) {
                        throw new SAXParseException("<span> inside glossary.",
                                locator);
                    } else if ("dt".equals(localName)) {
                        state = ParseState.IN_GLOSSARY_TERM;
                        buffer.setLength(0);
                        prevWasWhitespace = true;
                    } else if ("dd".equals(localName)) {
                        if (term == null) {
                            throw new SAXParseException(
                                    "Definition before term.", locator);
                        }
                        state = ParseState.IN_GLOSSARY_DEFINITION;
                    }
                    break;
                case IN_GLOSSARY_TERM:
                    if ("dl".equals(localName)) {
                        throw new SAXParseException(
                                "<dl> inside glossary term.", locator);
                    } else if ("span".equals(localName)) {
                        spanEndsToIgnore++;
                    } else if ("dt".equals(localName)) {
                        throw new SAXParseException(
                                "<dt> inside glossary term.", locator);
                    } else if ("dd".equals(localName)) {
                        throw new SAXParseException(
                                "<dd> inside glossary term.", locator);
                    }
                    break;
                case IN_GLOSSARY_DEFINITION:
                    if ("dl".equals(localName)) {
                        throw new SAXParseException(
                                "<dl> inside glossary definition.", locator);
                    } else if ("span".equals(localName)) {
                        state = ParseState.IN_GLOSSARY_DEFINITION_SPAN;
                        buffer.setLength(0);
                        prevWasWhitespace = true;
                    } else if ("dt".equals(localName)) {
                        throw new SAXParseException(
                                "<dt> inside glossary definition.", locator);
                    } else if ("dd".equals(localName)) {
                        throw new SAXParseException(
                                "<dd> inside glossary definition.", locator);
                    }
                    break;
                case IN_GLOSSARY_DEFINITION_SPAN:
                    if ("dl".equals(localName)) {
                        throw new SAXParseException(
                                "<dl> inside glossary definition span.",
                                locator);
                    } else if ("span".equals(localName)) {
                        spanEndsToIgnore++;
                    } else if ("dt".equals(localName)) {
                        throw new SAXParseException(
                                "<dt> inside glossary definition span.",
                                locator);
                    } else if ("dd".equals(localName)) {
                        throw new SAXParseException(
                                "<dd> inside glossary definition span.",
                                locator);
                    }
                    break;
            }
        }
    }

    /**
     * @see org.xml.sax.helpers.DefaultHandler#endElement(java.lang.String,
     *      java.lang.String, java.lang.String)
     */
    @Override
    public void endElement(String uri, String localName, String qName)
            throws SAXException {
        if ("http://www.w3.org/1999/xhtml".equals(uri)) {
            switch (state) {
                case OUTSIDE_GLOSSARY:
                    break;
                case IN_GLOSSARY:
                    if ("dl".equals(localName)) {
                        state = ParseState.OUTSIDE_GLOSSARY;
                    }
                    break;
                case IN_GLOSSARY_TERM:
                    if ("dt".equals(localName)) {
                        int len = buffer.length() - 1;
                        if (len >= 0 && buffer.charAt(len) == ' ') {
                            buffer.setLength(len);
                        }
                        term = buffer.toString();
                        state = ParseState.IN_GLOSSARY;
                    }
                    break;
                case IN_GLOSSARY_DEFINITION:
                    state = ParseState.IN_GLOSSARY;
                    break;
                case IN_GLOSSARY_DEFINITION_SPAN:
                    if (spanEndsToIgnore > 0) {
                        spanEndsToIgnore--;
                    } else {
                        int len = buffer.length() - 1;
                        if (len >= 0 && buffer.charAt(len) == ' ') {
                            buffer.setLength(len);
                        }
                        definitions.put(term, buffer.toString());
                        state = ParseState.IN_GLOSSARY_DEFINITION;
                    }
                    break;
            }
        }
    }

    /**
     * @see org.xml.sax.helpers.DefaultHandler#characters(char[], int, int)
     */
    @Override
    public void characters(char[] ch, int start, int length)
            throws SAXException {
        if (state == ParseState.IN_GLOSSARY_TERM
                || state == ParseState.IN_GLOSSARY_DEFINITION_SPAN) {
            for (int i = start; i < start + length; i++) {
                char c = ch[i];
                if (c == ' ' || c == '\t' || c == '\n' || c == '\r') {
                    if (!prevWasWhitespace) {
                        buffer.append(' ');
                        prevWasWhitespace = true;
                    }
                } else {
                    buffer.append(c);
                    prevWasWhitespace = false;
                }
            }
        }
    }

    /**
     * @see org.xml.sax.helpers.DefaultHandler#setDocumentLocator(org.xml.sax.Locator)
     */
    @Override
    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    /**
     * Returns the definitions.
     * 
     * @return the definitions
     */
    public Map<String, String> getDefinitions() {
        return definitions;
    }

}
