/*
 * Copyright (c) 2007 Henri Sivonen
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

package fi.iki.hsivonen.bib4ht;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Properties;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import net.sourceforge.texlipse.bibparser.BibParser;

import org.apache.xml.serializer.Method;
import org.apache.xml.serializer.OutputPropertiesFactory;
import org.apache.xml.serializer.Serializer;
import org.apache.xml.serializer.SerializerFactory;
import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import fi.iki.hsivonen.xml.CombineContentHandler;
import fi.iki.hsivonen.xml.HtmlSerializer;
import fi.iki.hsivonen.xml.SystemErrErrorHandler;

public class Main {
    public static void main(String[] args) throws IOException, SAXException,
            ParserConfigurationException {
        
        String src = args[0];
        String bib = args[1];
        String out = args [2];
        
        // set up error handler
        ErrorHandler eh = new SystemErrErrorHandler();

        // set up parser
        SAXParserFactory factory = SAXParserFactory.newInstance();
        factory.setNamespaceAware(true);
        factory.setValidating(false);
        XMLReader parser = factory.newSAXParser().getXMLReader();
        parser.setErrorHandler(eh);
        
        // setup bib parsing
        BibParser bibParser = new BibParser(new InputStreamReader(new FileInputStream(bib + ".bib"), "utf-8"));
        
        // gather toc and definitions
        TocGatherer tocGatherer = new TocGatherer();
        DefinitionGatherer definitionGatherer = new DefinitionGatherer();
        parser.setContentHandler(new CombineContentHandler(tocGatherer, definitionGatherer));
        parser.parse(new InputSource(new FileInputStream(src + ".xhtml")));
        
        // layer filters
        ParagraphIdFilter paraIdFilter = new ParagraphIdFilter(parser);
        AbbrFilter abbrFilter = new AbbrFilter(paraIdFilter, definitionGatherer.getDefinitions());
        TocFilter tocFilter = new TocFilter(abbrFilter, tocGatherer.getRoot());   
        DateFilter dateFilter = new DateFilter(tocFilter);        
        ReferenceFilter referenceFilter = "xtech".equals(src) ? new XTechReferenceFilter(dateFilter, bibParser.getEntries()) : new ReferenceFilter(dateFilter, bibParser.getEntries());
        
        // set up xml serializer
        
        Properties props = OutputPropertiesFactory.getDefaultMethodProperties(Method.XML);
        Serializer ser = SerializerFactory.getSerializer(props);
        ser.setOutputStream(new FileOutputStream(out + ".xhtml"));
        ContentHandler xmlSerializer = ser.asContentHandler();

        // set up html serializer
        
        ContentHandler htmlSerializer = new HtmlSerializer(new FileOutputStream(out + ".html"), HtmlSerializer.DOCTYPE_HTML401_STRICT, true);
        htmlSerializer = new SvgFiguresToPngContentHandler(htmlSerializer, new File("."));
        
        ContentHandler xmlLinkifier;
        ContentHandler htmlLinkifier;
        if ("xtech".equals(src)) {
            xmlLinkifier = xmlSerializer;
            htmlLinkifier = htmlSerializer;
        } else {
            xmlLinkifier = new LinkifyingContentHandler(xmlSerializer);
            htmlLinkifier = new LinkifyingContentHandler(htmlSerializer);
            ((LinkifyingContentHandler)xmlLinkifier).addLink("alternate", "application/pdf", "PDF version", out + ".pdf");
            ((LinkifyingContentHandler)xmlLinkifier).addLink("alternate", "text/html", "HTML version", out + ".html");
            ((LinkifyingContentHandler)htmlLinkifier).addLink("alternate", "application/pdf", "PDF version", out + ".pdf");
            ((LinkifyingContentHandler)htmlLinkifier).addLink("alternate", "application/xhtml+xml", "XHTML+SVG version", out + ".xhtml");
        }
        
        // parse
        referenceFilter.setProperty("http://xml.org/sax/properties/lexical-handler", xmlSerializer);
//        referenceFilter.setContentHandler(new Obfuscator(new CombineContentHandler(xmlSerializer, htmlSerializer)));
        referenceFilter.setContentHandler(new CombineContentHandler(xmlLinkifier, htmlLinkifier));
        referenceFilter.setErrorHandler(eh);
        referenceFilter.parse(new InputSource(new FileInputStream(src + ".xhtml")));
    }
}
