/*
 * Copyright (c) 2007 Henri Sivonen
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

package fi.iki.hsivonen.bib4ht;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import net.sourceforge.texlipse.bibparser.BibEntry;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLFilterImpl;

import fi.iki.hsivonen.xml.AttributesImpl;
import fi.iki.hsivonen.xml.XhtmlSaxEmitter;

public class ReferenceFilter extends XMLFilterImpl {

    private final static char[] LEFT = {'['};

    private final static char[] RIGHT = {']'};

    private final static char[] COMMA = {',', ' '};

    private final static char[] PERIOD = {'.', ' '};
    
    private final static String NON_XHTML = "\u0000";
    
    private XhtmlSaxEmitter emitter;
    
    private final char[] buffer = new char[32];
    
    private int pos = -1;
    
    private LinkedList<String> stack = new LinkedList<String>(); // grows from head
    
    private final Map<String, BibEntry> entries = new HashMap<String, BibEntry>();

    private final SortedSet<BibEntry> usedEntries = new TreeSet<BibEntry>();
    
    private Locator locator;
    
    private final AttributesImpl attrs = new AttributesImpl();
    
    public ReferenceFilter(XMLReader parent, Collection<BibEntry> entries) {
        super(parent);
        for (BibEntry entry : entries) {
            this.entries.put(entry.getKey(), entry);
        }
    }
    
    private void push(String elt) {
        stack.addFirst(elt);
    }
    
    private void pop() {
        stack.removeFirst();
    }
    
    /**
     * @see org.xml.sax.helpers.XMLFilterImpl#characters(char[], int, int)
     */
    @Override
    public void characters(char[] ch, int start, int length) throws SAXException {
        if (!searchCharacters()) {
            super.characters(ch, start, length); 
            return;
        }
        int mark = start;
        int end = start + length;
        for (int i = start; i < end; i++) {
            char c = ch[i];
            if (pos == -1) {
                if (c == '[') {
                    if (mark < i) {
                        super.characters(ch, mark, i - mark);                                    
                    }
                    pos = 0;
                }
            } else {
                if (c == ']') {
                    if (pos == 0) {
                        emitter.characters(LEFT);
                        emitter.characters(RIGHT);
                    } else {
                        handleBuffer();
                    }
                    pos = -1;
                    mark = i + 1;
                } else if (pos == buffer.length || !isRefChar(c)) {
                    emitter.characters(LEFT);
                    super.characters(buffer, 0, pos);
                    pos = -1;
                    mark = i;
                } else {
                    buffer[pos] = c;
                    pos++;
                }
            }
        }
        if (pos == -1 && mark < end) {
            super.characters(ch, mark, end - mark);            
        }
    }

    private boolean searchCharacters() {
        if (stack.getFirst() == NON_XHTML) {
            return false;
        }
        for (String elt : stack) {
            if ("code".equals(elt)) {
                return false;
            }
        }
        return true;
    }



    private boolean isRefChar(char c) {
        return ((c >= '-' && c <= ':') || (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || (c == '+') || (c == '_'));
    }

    private void handleBuffer() throws SAXException {
        String key = String.copyValueOf(buffer, 0, pos);
        BibEntry entry = entries.get(key);
        if (entry == null) {
            super.warning(new SAXParseException("No entry for bib key "+key, locator));
            emitter.characters(LEFT);
            super.characters(buffer, 0, pos);
            emitter.characters(RIGHT);
        } else {
            usedEntries.add(entry);
            
            emitLinkedSymbol(entry);
        }
    }

    /**
     * @param key
     * @param entry
     * @throws SAXException
     */
    private void emitLinkedSymbol(BibEntry entry) throws SAXException {
        attrs.clear();
        attrs.addAttribute("href", "#"+toRefStr(entry.getKey()));
        emitter.startElement("a", attrs);
        emitter.characters(LEFT);
        emitter.characters(entry.getSymbol());
        emitter.characters(RIGHT);
        emitter.endElement("a");
    }

    private String toRefStr(String key) {
        return "ref-" + key;
    }

    /**
     * @see org.xml.sax.helpers.XMLFilterImpl#endElement(java.lang.String, java.lang.String, java.lang.String)
     */
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        pop();
        super.endElement(uri, localName, qName);
    }
    
    /**
     * @see org.xml.sax.helpers.XMLFilterImpl#setDocumentLocator(org.xml.sax.Locator)
     */
    @Override
    public void setDocumentLocator(Locator arg0) {
        this.locator = arg0;
        super.setDocumentLocator(arg0);
    }

    /**
     * @see org.xml.sax.helpers.XMLFilterImpl#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
     */
    @Override
    public void startElement(String uri, String localName, String qName,
            Attributes atts) throws SAXException {
        if (pos != -1) {
            emitter.characters(LEFT);
            super.characters(buffer, 0, pos);
            pos = -1;            
        }
        if ("http://www.w3.org/1999/xhtml".equals(uri)) {
            push(localName);
        } else {
            push(NON_XHTML);
        }
        super.startElement(uri, localName, qName,
                atts);
    }

    /**
     * @see org.xml.sax.helpers.XMLFilterImpl#setContentHandler(org.xml.sax.ContentHandler)
     */
    @Override
    public void setContentHandler(ContentHandler handler) {
        this.emitter = new XhtmlSaxEmitter(handler);
        super.setContentHandler(handler);
    }

    /**
     * @see org.xml.sax.helpers.XMLFilterImpl#processingInstruction(java.lang.String, java.lang.String)
     */
    @Override
    public void processingInstruction(String target, String data) throws SAXException {
        if ("references".equals(target)) {
            includeXrefs();
            emitReferences();
            dumpUnused();
        } else {
            super.processingInstruction(target, data);
        }
    }

    private void includeXrefs() {
        Set<BibEntry> xrefs = new HashSet<BibEntry>();
        for (BibEntry entry : usedEntries) {
            if (entry.hasField("crossref")) {
                String xrefKey = entry.getFieldValue("crossref");
                BibEntry xref = entries.get(xrefKey);
                if (xref == null) {
                    System.err.println("Dangling crossreference: " + xrefKey);
                } else {
                    xrefs.add(xref);
                }
            }
        }
        usedEntries.addAll(xrefs);
    }

    private void dumpUnused() {
        System.err.println("Unused entries:");
        for (Map.Entry<String, BibEntry> mapEntry : entries.entrySet()) {
            BibEntry entry = mapEntry.getValue();
            if (!usedEntries.contains(entry)) {
                System.err.println(entry.getKey());
            }
        }
    }

    private void emitReferences() throws SAXException {
        emitter.startElementWithClass("dl", "references");
        
        for (BibEntry entry : usedEntries) {
            emitEntry(entry);
        }
        
        emitter.endElement("dl");
    }

    private void emitEntry(BibEntry entry) throws SAXException {
        attrs.clear();
        attrs.addAttribute("id", toRefStr(entry.getKey()));
        emitter.startElement("dt", attrs);
        emitter.characters(LEFT);
        emitter.characters(entry.getSymbol());
        emitter.characters(RIGHT);        
        emitter.endElement("dt");

        emitter.startElement("dd");
        emitter.startElement("p");
        emitEntryContents(entry);
        emitter.endElement("p");
        emitNote(entry);
        emitter.endElement("dd");        
    }

    private void emitNote(BibEntry entry) throws SAXException {
        if (entry.hasField("note")) {
            emitter.startElementWithClass("p", "note");
            emitter.characters(entry.getFieldValue("note"));
            emitter.endElement("p"); 
        }
    }

    private void emitEntryContents(BibEntry entry) throws SAXException {
        String url = entry.getFieldValue("url");
        if (url == null) {
            url = entry.getFieldValue("doi");
        }
        
        if (emitIfPresent(entry, "stdnumber", "span")) {
            emitter.characters(COMMA);
        }

        if ("private".equals(entry.getType())) {
            emitter.characters("Private communication with ");
        } else {
            String title = entry.getFieldValue("title");
            if (title == null) {
                error(new SAXParseException("Bib entry without a title "
                        + entry.getKey(), locator));
            } else {
                if (url != null) {
                    attrs.clear();
                    attrs.addAttribute("href", url);
                    emitter.startElement("a", attrs);
                }
                emitField(title, "title", "cite");
                if (url != null) {
                    emitter.endElement("a");
                }
                // XXX Working Draft!!
                emitWD(entry);
                emitter.characters(PERIOD);
            }
        }
        
        if (entry.hasField("author")) {
            emitter.startElementWithClass("span", "author");
            emitNames(entry.getFieldValue("author"));
            emitter.endElement("span");            
            emitter.characters(PERIOD);
        }

        if (entry.hasField("editor")) {
            emitter.startElementWithClass("span", "editor");
            boolean plural = emitNames(entry.getFieldValue("editor"));
            emitter.endElement("span");
            if (plural) {
                emitter.characters(", editors. ");
            } else {
                emitter.characters(", editor. ");                
            }
        }
        
        if (entry.hasField("booktitle") || entry.hasField("journal")) {
            emitter.characters("In ");
            emitIfPresent(entry, "booktitle", "cite");
            emitIfPresent(entry, "journal", "cite");

            emitLabeledFieldIfPresent(entry, "series");
            emitLabeledFieldIfPresent(entry, "volume");
            emitLabeledFieldIfPresent(entry, "number");
            emitLabeledFieldIfPresent(entry, "pages");
            
            emitter.characters(PERIOD);            
        }
 
        emitThesis(entry);
        
        boolean hadOrg = false;
        if (entry.hasField("school")) {
            hadOrg = true;
            emitIfPresent(entry, "school", "span");
        } else if (entry.hasField("organization")) {
            hadOrg = true;
            emitIfPresent(entry, "organization", "span");
        } else if (entry.hasField("publisher")) {
            hadOrg = true;
            emitIfPresent(entry, "publisher", "span");
        }
        
        if (entry.hasField("year")) {
            if (hadOrg) {
                emitter.characters(COMMA);
            }
            emitIfPresent(entry, "year", "span");
            emitter.characters(PERIOD);
        } else {
            if (hadOrg) {
                emitter.characters(PERIOD);
            }
        }
        
        if (entry.hasField("isbn")) {
            emitter.characters("ISBN:\u00A0");
            emitIfPresent(entry, "isbn", "span");
            emitter.characters(PERIOD);
        }
        
        if (entry.hasField("issn")) {
            emitter.characters("ISSN:\u00A0");
            emitIfPresent(entry, "issn", "span");
            emitter.characters(PERIOD);
        }
        
        if (entry.hasField("crossref")) {
            BibEntry xref = entries.get(entry.getFieldValue("crossref"));
            if (xref != null) {
                emitter.characters("Published in ");
                emitLinkedSymbol(xref);
                emitter.characters(PERIOD);
            }
        }
        
        if (url != null) {
            emitter.startElementWithClass("span", "urlwrap");
            
            attrs.clear();
            attrs.addAttribute("href", url);
            attrs.addAttribute("class", "url");
            emitter.startElement("a", attrs);
            
            // w3.org is known to redirect to www.w3.org, so let's 
            // avoid line wrapping
            
            //if (url.startsWith("http://www.w3.org/")) {
            //    emitter.characters("http://w3.org/" + url.substring(18));
            //} else {
                emitter.characters(url);
            //}
            
            emitter.endElement("a");
            
            if (entry.hasField("refdate")) {
                emitter.characters(" (referenced:\u00A0");
                emitIfPresent(entry, "refdate", "span");
                emitter.characters(")");
            }
            
            emitter.endElement("span");
        }
    }

    private void emitThesis(BibEntry entry) throws SAXException {
        if ("mastersthesis".equals(entry.getType())) {
            emitter.characters("Master\u2019s thesis, ");
        } else if ("phdthesis".equals(entry.getType())) {
            emitter.characters("PhD thesis, ");
        } else if ("bachelorsthesis".equals(entry.getType())) {
            emitter.characters("Bachelor\u2019s thesis, ");
        }
    }

    private void emitWD(BibEntry entry) throws SAXException {
        if ("workingdraft".equals(entry.getType())) {
            emitter.characters(", working draft");
        }   
    }

    private boolean emitNames(String fieldValue) throws SAXException {
        String[] arr = fieldValue.split("\\s+and\\s+");
        boolean emitted = false;
        for (int i = 0; i < arr.length; i++) {
            String name = arr[i];
            if (emitted) {
                if (i == arr.length - 1) {
                    emitter.characters(" and ");
                } else {
                    emitter.characters(COMMA);
                }
            }
            emitField(name, "fn", "span");
            emitted = true;
        }
        return arr.length != 1;
    }

    /**
     * @param entry
     * @throws SAXException
     */
    private void emitLabeledFieldIfPresent(BibEntry entry, String fieldName) throws SAXException {
        if (entry.hasField(fieldName)) {
            emitter.characters(", "+fieldName+" ");
            emitField(entry.getFieldValue(fieldName), fieldName, "span");
        }
    }
    
    private boolean emitIfPresent(BibEntry entry, String fieldName, String eltName) throws SAXException {
        if (entry.hasField(fieldName)) {
            emitField(entry.getFieldValue(fieldName), fieldName, eltName);
            return true;
        } else {
            return false;
        }
    }

    private void emitField(String fieldValue, String fieldName, String eltName) throws SAXException {
        emitter.startElementWithClass(eltName, fieldName);
        emitter.characters(fieldValue);
        emitter.endElement(eltName);
    }


}
