/*
 * Copyright (c) 2007 Henri Sivonen
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

package fi.iki.hsivonen.bib4ht;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.batik.transcoder.TranscoderException;
import org.apache.batik.transcoder.TranscoderInput;
import org.apache.batik.transcoder.TranscoderOutput;
import org.apache.batik.transcoder.image.PNGTranscoder;
import org.apache.xml.serializer.Method;
import org.apache.xml.serializer.OutputPropertiesFactory;
import org.apache.xml.serializer.Serializer;
import org.apache.xml.serializer.SerializerFactory;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import fi.iki.hsivonen.xml.AttributesImpl;

public class SvgFiguresToPngContentHandler implements ContentHandler {

    private static final float PIXEL_FACTOR = 2.0f;

    private static final float IMG_FACTOR = 0.5f;
    
    private static final Pattern FIGURE = Pattern.compile(".*\\bfigure\\b.*");

    private static final Pattern VIEW_BOX = Pattern.compile("\\s*\\S+\\s+\\S+\\s+(\\S+)\\s+(\\S+)\\s*");

    private final AttributesImpl attrs = new AttributesImpl();
    
    private File parentDir;
    
    private ContentHandler next;

    private ContentHandler delegate;

    private int figureDepth = 0;

    private int svgDepth = 0;

    private int descDepth = 0;

    private final StringBuilder descBuffer = new StringBuilder();

    private float width;

    private float height;

    private boolean prevWasWhitespace;

    private File tempFile;

    private String desc = "";

    private String id;
    
    private Locator locator;

    public SvgFiguresToPngContentHandler(ContentHandler next, File parentDir) {
        this.next = next;
        this.delegate = next;
        this.parentDir = parentDir;
    }
    
    /**
     * @param ch
     * @param start
     * @param length
     * @throws SAXException
     * @see org.xml.sax.ContentHandler#characters(char[], int, int)
     */
    public void characters(char[] ch, int start, int length)
            throws SAXException {
        if (descDepth > 0) {
            for (int i = start; i < start + length; i++) {
                char c = ch[i];
                if (c == ' ' || c == '\t' || c == '\n' || c == '\r') {
                    if (!prevWasWhitespace) {
                        descBuffer.append(' ');
                        prevWasWhitespace = true;
                    }
                } else {
                    descBuffer.append(c);
                    prevWasWhitespace = false;
                }
            }

        }
        delegate.characters(ch, start, length);
    }

    /**
     * @throws SAXException
     * @see org.xml.sax.ContentHandler#endDocument()
     */
    public void endDocument() throws SAXException {
        if (tempFile != null) {
            tempFile.delete();
            tempFile = null;
        }
        delegate.endDocument();
    }

    /**
     * @param uri
     * @param localName
     * @param qName
     * @throws SAXException
     * @see org.xml.sax.ContentHandler#endElement(java.lang.String,
     *      java.lang.String, java.lang.String)
     */
    public void endElement(String uri, String localName, String qName)
            throws SAXException {
        delegate.endElement(uri, localName, qName);

        if (descDepth > 0) {
            descDepth--;
            if (descDepth == 0) {
                int len = descBuffer.length() - 1;
                if (len >= 0 && descBuffer.charAt(len) == ' ') {
                    descBuffer.setLength(len);
                }
                desc = descBuffer.toString();
            }
        }

        if (figureDepth > 0) {
            figureDepth--;
            if (figureDepth == 0) {
                id = null;
            }
        }

        if (svgDepth > 0) {
            svgDepth--;
            if (svgDepth == 0) {
                delegate.endDocument();
                delegate = next;

                String imgName = id + ".png";
                
                try {
                    PNGTranscoder t = new PNGTranscoder();
                    t.addTranscodingHint(PNGTranscoder.KEY_WIDTH, new Float(
                            width * PIXEL_FACTOR));
                    t.addTranscodingHint(PNGTranscoder.KEY_HEIGHT, new Float(
                            height * PIXEL_FACTOR));
                    TranscoderInput input = new TranscoderInput(
                            new FileInputStream(tempFile));
                    OutputStream ostream = new FileOutputStream(new File(parentDir, imgName));
                    TranscoderOutput output = new TranscoderOutput(ostream);
                    t.transcode(input, output);
                    ostream.flush();
                    ostream.close();
                } catch (IOException e) {
                    throw new SAXException(e);
                } catch (TranscoderException e) {
                    throw new SAXException(e);
                }

                attrs.clear();
                attrs.addAttribute("src", imgName);
                attrs.addAttribute("alt", desc);
                attrs.addAttribute("width", Integer.toString(Math.round(width * IMG_FACTOR)));
                attrs.addAttribute("height", Integer.toString(Math.round(height * IMG_FACTOR)));
                delegate.startElement("http://www.w3.org/1999/xhtml", "img", "img", attrs);
                delegate.endElement("http://www.w3.org/1999/xhtml", "img", "img");                
            }
        }
    }

    /**
     * @param prefix
     * @throws SAXException
     * @see org.xml.sax.ContentHandler#endPrefixMapping(java.lang.String)
     */
    public void endPrefixMapping(String prefix) throws SAXException {
        delegate.endPrefixMapping(prefix);
    }

    /**
     * @param ch
     * @param start
     * @param length
     * @throws SAXException
     * @see org.xml.sax.ContentHandler#ignorableWhitespace(char[], int, int)
     */
    public void ignorableWhitespace(char[] ch, int start, int length)
            throws SAXException {
        delegate.ignorableWhitespace(ch, start, length);
    }

    /**
     * @param target
     * @param data
     * @throws SAXException
     * @see org.xml.sax.ContentHandler#processingInstruction(java.lang.String,
     *      java.lang.String)
     */
    public void processingInstruction(String target, String data)
            throws SAXException {
        delegate.processingInstruction(target, data);
    }

    /**
     * @param locator
     * @see org.xml.sax.ContentHandler#setDocumentLocator(org.xml.sax.Locator)
     */
    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
        delegate.setDocumentLocator(locator);
    }

    /**
     * @param name
     * @throws SAXException
     * @see org.xml.sax.ContentHandler#skippedEntity(java.lang.String)
     */
    public void skippedEntity(String name) throws SAXException {
        delegate.skippedEntity(name);
    }

    /**
     * @throws SAXException
     * @see org.xml.sax.ContentHandler#startDocument()
     */
    public void startDocument() throws SAXException {
        figureDepth = 0;
        svgDepth = 0;
        descDepth = 0;
        descBuffer.setLength(0);
        desc = "";
        id = null;
        try {
            tempFile = File.createTempFile("figure", ".svg");
        } catch (IOException e) {
            throw new SAXException(e);
        }

        delegate.startDocument();
    }

    /**
     * @param uri
     * @param localName
     * @param qName
     * @param atts
     * @throws SAXException
     * @see org.xml.sax.ContentHandler#startElement(java.lang.String,
     *      java.lang.String, java.lang.String, org.xml.sax.Attributes)
     */
    public void startElement(String uri, String localName, String qName,
            Attributes atts) throws SAXException {
        if (figureDepth > 0) {
            figureDepth++;

            if (svgDepth > 0) {
                svgDepth++;

                if (descDepth > 0) {
                    descDepth++;
                } else if ("http://www.w3.org/2000/svg".equals(uri)
                        && "desc".equals(localName)) {
                    descDepth = 1;
                    prevWasWhitespace = true;
                    descBuffer.setLength(0);
                }

            } else if ("http://www.w3.org/2000/svg".equals(uri)
                    && "svg".equals(localName)) {
                svgDepth = 1;
                desc = "";
                String viewBox = atts.getValue("", "viewBox");
                if (viewBox != null) {
                    Matcher m = VIEW_BOX.matcher(viewBox);
                    if (m.matches()) {
                        String widthStr = m.group(1);
                        String heightStr = m.group(2);
                        try {
                            width = Float.parseFloat(widthStr);
                            height = Float.parseFloat(heightStr);
                        } catch (NumberFormatException e) {
                            throw new SAXParseException(e.getMessage(), locator);
                        }
                    } else {
                        throw new SAXParseException("Bad viewBox syntax.", locator);
                    }
                } else {
                    throw new SAXParseException("No viewBox.", locator);
                }

                Properties props = OutputPropertiesFactory.getDefaultMethodProperties(Method.XML);
                Serializer ser = SerializerFactory.getSerializer(props);
                try {
                    ser.setOutputStream(new FileOutputStream(tempFile));
                    delegate = ser.asContentHandler();
                } catch (IOException e) {
                    throw new SAXException(e);
                }

                delegate.startDocument();
            }

        } else {
            String klazz = atts.getValue("", "class");
            if (klazz != null) {
                Matcher m = FIGURE.matcher(klazz);
                if (m.matches()) {
                    figureDepth = 1;
                    id = atts.getValue("", "id");
                    if (id == null) {
                        throw new SAXParseException("No id for figure.", locator);
                    }
                }
            }
        }
        delegate.startElement(uri, localName, qName, atts);
    }

    /**
     * @param prefix
     * @param uri
     * @throws SAXException
     * @see org.xml.sax.ContentHandler#startPrefixMapping(java.lang.String,
     *      java.lang.String)
     */
    public void startPrefixMapping(String prefix, String uri)
            throws SAXException {
        delegate.startPrefixMapping(prefix, uri);
    }
}
