/*
 * $Id: BibParser.java,v 1.1 2007/01/25 19:22:54 hsivonen Exp $
 *
 * Copyright (c) 2004-2005 by the TeXlapse Team.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
package net.sourceforge.texlipse.bibparser;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PushbackReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.List;

import net.sourceforge.texlipse.bibparser.lexer.LexerException;
import net.sourceforge.texlipse.bibparser.node.Start;
import net.sourceforge.texlipse.bibparser.parser.Parser;
import net.sourceforge.texlipse.bibparser.parser.ParserException;


/**
 * BibTeX parser front-end. After creation, the parsing is done by calling
 * the getEntries() -method, after which getAbbrevs() and getErrors() should
 * be called (otherwise the data returned by these two is essentially meaningless.)
 * 
 * @author Oskar Ojala
 */
public class BibParser {
    
    private String filename;
    private Reader reader;
    
    private List<String> warnings = new ArrayList<String>();
    private Start ast;
    
    /**
     * Constructs a new BibTeX parser.
     * 
     * @param filename The file to parse
     */
    public BibParser(String filename) {
        this.filename = filename;
    }
    
    /**
     * Constructs a new BibTeX parser.
     * 
     * @param r A reader to the BibTeX-data to parse
     */
    public BibParser(Reader r) {
        this.reader = r;
    }
    
    /**
     * Parses the document, constructs a list of the entries and returns
     * them.
     * 
     * @return BibTeX entries (<code>ReferenceEntry</code>)
     */
    public List<BibEntry> getEntries() throws IOException, FileNotFoundException {
        try {
            BibLexer l;
            if (filename != null) {
                l = new BibLexer(new PushbackReader (new FileReader(filename), 1024));
            } else {
                l = new BibLexer(new PushbackReader(reader, 1024));
            }
            
            Parser p = new Parser(l);
            this.ast = p.parse();
            
//            AbbrevRetriever ar = new AbbrevRetriever();
//            ast.apply(ar);
//            Map<String,String> abbrevs = ar.getAbbrevs();
            
            // Then make the datastructs
            EntryRetriever er = new EntryRetriever();
            ast.apply(er);
            warnings = er.getWarnings();
            
            return er.getEntries();
            
            // TODO some fancy error logging could be added...
        } catch (LexerException le) {
            System.err.println("LexingError: " + le.getMessage());
        } catch (ParserException pe) {
            System.err.println("ParseError: " + pe.getMessage());
        }
        return null;
    }
        
    /**
     * @return Returns the warnings.
     */
    public List<String> getWarnings() {
        return warnings;
    }
    
}
