/*
 * $Id: EntryRetriever.java,v 1.1 2007/01/25 19:22:54 hsivonen Exp $
 *
 * Copyright (c) 2004-2005 by the TeXlapse Team.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
package net.sourceforge.texlipse.bibparser;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sourceforge.texlipse.bibparser.analysis.DepthFirstAdapter;
import net.sourceforge.texlipse.bibparser.node.ABibeBibEntry;
import net.sourceforge.texlipse.bibparser.node.ABibstreBibEntry;
import net.sourceforge.texlipse.bibparser.node.ABibtaskBibEntry;
import net.sourceforge.texlipse.bibparser.node.ABibtex;
import net.sourceforge.texlipse.bibparser.node.AConcat;
import net.sourceforge.texlipse.bibparser.node.AEntryDef;
import net.sourceforge.texlipse.bibparser.node.AEntrybraceEntry;
import net.sourceforge.texlipse.bibparser.node.AEntryparenEntry;
import net.sourceforge.texlipse.bibparser.node.AIdValOrSid;
import net.sourceforge.texlipse.bibparser.node.AKeyvalDecl;
import net.sourceforge.texlipse.bibparser.node.ANumValOrSid;
import net.sourceforge.texlipse.bibparser.node.AStrbraceStringEntry;
import net.sourceforge.texlipse.bibparser.node.AStrparenStringEntry;
import net.sourceforge.texlipse.bibparser.node.AValueBValOrSid;
import net.sourceforge.texlipse.bibparser.node.AValueQValOrSid;
import net.sourceforge.texlipse.bibparser.node.TIdentifier;
import net.sourceforge.texlipse.bibparser.node.TStringLiteral;
import net.sourceforge.texlipse.bibparser.node.Token;


/**
 * Retrieves the BibTeX entries from the AST.
 * 
 * This class is a visitor, that is applied on the AST that is a result of parsing a
 * BibTeX-file. See <a href="http://www.sablecc.org">http://www.sablecc.org</a> for
 * more information on the structure of the AST and the visitors.
 * 
 * @author Oskar Ojala
 */
public final class EntryRetriever extends DepthFirstAdapter {

    // for cross referencing
    private final class EntryText {
        Token token;
        Set definedFields;
        public EntryText(Token t, Set df) {
            token = t;
            definedFields = df;
        }
    }
    
    //private ArrayList<ParseErrorMessage> warnings = new ArrayList<ParseErrorMessage>();
    private List<String> warnings = new ArrayList<String>();
   
    private List tasks = new ArrayList(); // type: ParseErrorMessage
    
    private List<BibEntry> entries = new ArrayList<BibEntry>();
    
    private BibEntry currEntry;
    private StringBuilder currFieldValue;
    private String currField;
    private String crossref;

    /**
     * Currently defined fields for an entry
     */
    private Set<String> currDefinedFields = new HashSet<String>();

    /**
     * All defined keys -- can be used for testing whether a key is unique
     */
    private Set<String> allDefinedKeys = new HashSet<String>();;
    
    private static final Map<String,String> predefAbbrevs = new HashMap<String,String>();
//    private Map<String,List<String>> crossrefs = new HashMap<String,List<String>>();
    
    private Map<String,String> abbrevs;
    
    /**
     * A list of required fields for the different BibTeX entries
     */
    private static final Map<String, List<String>> requiredFieldsPerType = new HashMap<String, List<String>>();
    
    static {
        predefAbbrevs.put("jan", "January");
        predefAbbrevs.put("feb", "February");
        predefAbbrevs.put("mar", "March");
        predefAbbrevs.put("apr", "April");
        predefAbbrevs.put("may", "May");
        predefAbbrevs.put("jun", "June");
        predefAbbrevs.put("jul", "July");
        predefAbbrevs.put("aug", "August");
        predefAbbrevs.put("sep", "September");
        predefAbbrevs.put("oct", "October");
        predefAbbrevs.put("nov", "November");
        predefAbbrevs.put("dec", "December");

        String[] article = {"author", "title", "journal", "year"};                     
        String[] book = {"title", "publisher", "year"};
        String[] booklet = {"title"};
        String[] conference = {"author", "title", "booktitle", "year"};
        String[] inbook = {"title", "publisher", "year"};
        String[] incollection = {"author", "title", "booktitle", "publisher", "year"};
        String[] inproceedings = {"author", "title", "booktitle", "year"};;
        String[] manual = {"title"};
        String[] mastersthesis = {"author", "title", "school", "year"};
        String[] phdthesis = {"author", "title", "school", "year"};        
        String[] techreport = {"author", "title", "institution", "year"};
        String[] proceedings = {"title", "year"};
        String[] unpublished = {"author", "title", "note"};

        requiredFieldsPerType.put("article", new ArrayList<String>(Arrays.asList(article)));
        requiredFieldsPerType.put("book", new ArrayList<String>(Arrays.asList(book)));
        requiredFieldsPerType.put("booklet", new ArrayList<String>(Arrays.asList(booklet)));
        requiredFieldsPerType.put("conference", new ArrayList<String>(Arrays.asList(conference)));
        requiredFieldsPerType.put("inbook", new ArrayList<String>(Arrays.asList(inbook)));
        requiredFieldsPerType.put("incollection", new ArrayList<String>(Arrays.asList(incollection)));
        requiredFieldsPerType.put("inproceedings", new ArrayList<String>(Arrays.asList(inproceedings)));
        requiredFieldsPerType.put("manual", new ArrayList<String>(Arrays.asList(manual)));
        requiredFieldsPerType.put("mastersthesis", new ArrayList<String>(Arrays.asList(mastersthesis)));        
        requiredFieldsPerType.put("phdthesis", new ArrayList<String>(Arrays.asList(phdthesis)));
        requiredFieldsPerType.put("techreport", new ArrayList<String>(Arrays.asList(techreport)));
        requiredFieldsPerType.put("proceedings", new ArrayList<String>(Arrays.asList(proceedings)));
        requiredFieldsPerType.put("unpublished", new ArrayList<String>(Arrays.asList(unpublished)));

    }

    public EntryRetriever() {
        this.abbrevs = new HashMap<String,String>(predefAbbrevs);
    }
    
    /**
     * @return The entries as a list of <code>ReferenceEntry</code>s
     */
    public List<BibEntry> getEntries() {
        return entries;
    }
    
    /**
     * @return A list of warnings in the file
     */
    public List getWarnings() {
        return warnings;
    }
    
    /**
     * @return A list of task markers in the file
     */
    public List getTasks() {
        return tasks;
    }
    
    public void inABibtex(ABibtex node) {
    }
    
    public void outABibtex(ABibtex node) {
    }
    
    public void inABibeBibEntry(ABibeBibEntry node) {
    }

    public void inABibstreBibEntry(ABibstreBibEntry node) {
    }

    public void inABibtaskBibEntry(ABibtaskBibEntry node) {
    }

    public void outABibeBibEntry(ABibeBibEntry node) {
    }

    public void outABibstreBibEntry(ABibstreBibEntry node) {
    }

    public void outABibtaskBibEntry(ABibtaskBibEntry node) {
    }

    private void inAnAbbrev(TIdentifier tid, TStringLiteral tsl) {
        if (abbrevs.put(tid.getText(), tsl.getText()) != null) {
            System.err.println("String key " + tid.getText() + " is not unique, line " + tid.getLine());
        }
    }
    
    public void inAStrbraceStringEntry(AStrbraceStringEntry node) {
        inAnAbbrev(node.getIdentifier(), node.getStringLiteral());
    }
    
    public void outAStrbraceStringEntry(AStrbraceStringEntry node) {
    }
    
    public void inAStrparenStringEntry(AStrparenStringEntry node) {
        inAnAbbrev(node.getIdentifier(), node.getStringLiteral());
    }
    
    public void outAStrparenStringEntry(AStrparenStringEntry node) {
    }
    
    private void inBibtexEntry(TIdentifier tid) {
        crossref = null;
        currEntry = new BibEntry(tid.getText());
        if (!allDefinedKeys.add(tid.getText())) {
            System.err.println("BibTex key " + tid.getText() + " is not unique, line " + tid.getLine());
        }
    }

    private void checkMissingFields(String type, Set<String> fields) {
        List<String> reqFieldList = requiredFieldsPerType.get(type);
        if (reqFieldList != null) {
            for (String reqField : reqFieldList) {
                if (!fields.contains(reqField)) {
                    warnings.add("Entry " + currEntry.getKey() + " is missing required field " + reqField);
                }
            }
        } else {
            warnings.add("Unknown entry type " + type);
        }
    }
    
    private void outBibtexEntry() {
        entries.add(currEntry);
        checkMissingFields(currEntry.getType(), currDefinedFields);
        currDefinedFields.clear();
    }
    
    /**
     * Called when entering a bibliography entry, starts
     * forming an entry for the entry list
     *
     * @param node an <code>AEntry</code> value
     */
    public void inAEntrybraceEntry(AEntrybraceEntry node) {
        inBibtexEntry(node.getIdentifier());
    }
    
    /**
     * Called when exiting a bibliography entry, adds the formed
     * entry into the entry list
     *
     * @param node an <code>AEntry</code> value
     */
    public void outAEntrybraceEntry(AEntrybraceEntry node) {
        outBibtexEntry();
    }
    
    public void inAEntryparenEntry(AEntryparenEntry node) {
        inBibtexEntry(node.getIdentifier());
    }
    
    public void outAEntryparenEntry(AEntryparenEntry node) {
        outBibtexEntry();
    }
    
    public void inAEntryDef(AEntryDef node) {
    }
    
    /**
     * Handles the type of the bibliography entry
     *
     * @param node an <code>AEntryDef</code> value
     */
    public void outAEntryDef(AEntryDef node) {
        currEntry.setType(node.getEntryName().getText().substring(1).toLowerCase());
    }
    
    public void inAKeyvalDecl(AKeyvalDecl node) {
        currFieldValue = new StringBuilder();
        currField = node.getIdentifier().getText().toLowerCase();
        if (!currDefinedFields.add(currField)) {
            warnings.add("Field " + currField + " appears more than once in entry " + currEntry.getKey());
        }
    }
    
    public void outAKeyvalDecl(AKeyvalDecl node) {
        currEntry.setFieldValue(currField, currFieldValue.toString());
    }
    
    public void inAConcat(AConcat node) {
    }
    
    public void outAConcat(AConcat node) {
    }
    
    public void inAValueBValOrSid(AValueBValOrSid node) {
    }

    public void inAValueQValOrSid(AValueQValOrSid node) {
    }

    public void outAValueBValOrSid(AValueBValOrSid node) {
        outAValueValOrSid(node.getStringLiteral().getText(),
                node.getStringLiteral());
    }

    public void outAValueQValOrSid(AValueQValOrSid node) {
        TStringLiteral tsl = node.getStringLiteral();
        if (tsl != null) {
            outAValueValOrSid(tsl.getText(), tsl);
        } else {
            warnings.add(currField + " is empty in " + currEntry.getKey());
        }
    }
    
    private void outAValueValOrSid(String text, Token tsl) {
        String fieldValue = text.replaceAll("\\s+", " ");
        currFieldValue.append(fieldValue);

        if ("crossref".equals(currField)) {
            crossref = fieldValue;
        }
        // Test for empty fields
        if (fieldValue.equalsIgnoreCase("")) {
            warnings.add(currField + " is empty in " + currEntry.getKey());
        }
    }

    public void inANumValOrSid(ANumValOrSid node) {
    }

    public void outANumValOrSid(ANumValOrSid node) {
        outAValueValOrSid(node.getNumber().getText(), node.getNumber());
    }
    
    public void inAIdValOrSid(AIdValOrSid node) {
    }
    
    public void outAIdValOrSid(AIdValOrSid node) {
        TIdentifier tid = node.getIdentifier();
        String expansion = abbrevs.get(tid.getText());
        if (expansion != null) {
            outAValueValOrSid(expansion, tid);
        } else {
            warnings.add("The abbreviation " + tid.getText() + " is undefined");
        }
    }
}
